#
# mtable.rb
#
#   Copyright (c) 2001,2002 Minero Aoki <aamine@loveruby.net>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#

require 'refe/completion'


module ReFe

  class MethodTable

    def initialize
      @table = {}      # c. => {m=>true}
      @inv_table = {}  # m  => [c. ...]
      @c_comptable = CompletionTable.new
      @m_comptable = CompletionTable.new
    end

    def inspect
      "#<#{type}>"
    end

    #
    # register
    #

    def register_method( name )
      c, t, m = split_name(name)
      unless table = @table[c+t] then
        @table[c+t] = table = {}
        @c_comptable.add c
      end
      table[m] = true
      @m_comptable.add m
      @inv_table[m] = (@inv_table[m] || []) | [c+t]
    end

    def register_class( c )
      @c_comptable.add c
    end

    TYPE_RE = /[\.\#]/

    def split_name( name )
      c, m = name.split( TYPE_RE, 2 )
      t = TYPE_RE.match(name)[0]

      return c, t, m
    end
    private :split_name

    #
    # lookup
    #

    def registered_classes
      @c_comptable.candidates
    end

    def methods_of( c )
      @table[c] or return nil
      @table[c].keys
    end

    #
    # completion
    #

    def complete_class( klass, type )
      ret = @c_comptable.complete(klass).collect {|c|
          tmp = (type ? [type] : %w(. #)).collect {|t|
              @table.key?(c+t) ? c+t : nil
          }.compact
          (tmp.empty? and not type) ? [c] : tmp
      }.flatten
      ret.empty? and notfound 'class', klass

      ret
    end

    def complete_method( klass, type, method )
      cc = @c_comptable.complete( klass )
      ccand = (type ? [type] : %w(. #)).collect {|t|
          cc.collect {|c| @table.key?(c+t) ? c+t : nil }
      }.flatten.compact
      ccand.empty? and notfound 'class', klass

      mcand = @m_comptable.complete_method( method )
      mcand.empty? and notfound 'method', method

      ret = ccand.collect {|c| mcand.collect {|m|
              @table[c][m] ? c+m : nil
      }}.flatten.compact
      ret.empty? and notfound 'method', "#{klass}#{type ? type : ' '}#{method}"

      ret
    end

    def complete_method2( method )
      mcand = @m_comptable.complete_method( method )
      mcand.empty? and notfound 'method', method

      mcand.collect {|m|@inv_table[m].collect {|c|  c+m  }}.flatten
    end

    private

    def notfound( type, name )
      raise SearchError, "no such #{type}: #{name}"
    end

  end

end
