#
# rubysrcparser.rb
#
#   Copyright (c) 2001,2002 Minero Aoki <aamine@loveruby.net>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#

require 'refe/mtof'


module ReFe

  class RubySourceCodeParser

    ::ReFe.declear_parser self

    def self.process( dir, db, msgout )
      db.f_src_updating {
      db.mtof_updating {
          Dir.glob( dir + '/*.[cy]' ).each do |fname|
            msgout.print_filename fname
            new.parse fname, db

            next if File.basename(fname) == 'class.c'
            File.open( fname ) {|f|
                MtoF.new.parse f, db
            }
          end
      } }
    end


    def parse( fname, db )
      File.open( fname ) {|f|
        _parse f, fname, db
      }
    end

    def _parse( f, fname, db )
      pre = nil
      while line = f.gets do
        case line
        when /\A\w/
          (pre ||= []).push line
        when /\A\s*\z/
          pre = nil
        when /\A\{/
          if pre then
            db.register_fsrc get_fname(pre),
                             pre.join(''), read_function(f, line),
                             "#{fname}:#{f.lineno}"
          end
          pre = nil
        else
          pre.push line if pre
        end
      end
    end

    def get_fname( pre )
      pre.reverse_each do |line|
        /\A\w/ === line                        and
        not /NORETURN/ === line                and
        m = /(\w+)(?:\s*|\s+_)\(/.match(line)  and
        return m[1]
      end
      raise ArgumentError, "cannot get func name: #{pre}"
    end

    def read_function( f, line )
      ret = [line]
      while line = f.gets do
        ret.push line
        break if /\A\};?\s*\z/ === line
      end
      ret.join('')
    end
  
  end   # class FunctionIndexBuilder

end
